#include <utility>
#include <cstdio>
#include <vector>
#include <queue>
#include <map>
#include <fcntl.h>
#include <sys/poll.h>
#include <csignal>
#include <errno.h>
#include <unistd.h>

#include "buffer.hh"
#include "alloc.hh"

using std::make_pair;

unsigned long long MaxBufferSize = 100*1000000; // 100 MB

static unsigned long long total_bytes = 0;
static unsigned long long written_bytes = 0;

class buffer
{
    std::queue<BufferType> buffers;
    size_t frontptr;
public:
    buffer() : buffers(), frontptr(0)
    {
    }
    buffer(const BufferType& data) : buffers(), frontptr(0)
    {
        PushBack(data);
    }
    
    void PushBack(const BufferType& data)
    {
     /* Don't use string append - it's too slow and we don't need it. */
        buffers.push(data);
    }
    void PopFront(size_t n)
    {
        while(n > 0 && !buffers.empty())
        {
            size_t c = GetFrontSize();
            if(n < c)
            {
                frontptr += n;
                c = n; // ate "n" bytes
            }
            else
            {
                buffers.pop();
                frontptr = 0;
                // ate everything that was left in the front buf
            }
            n -= c;
        }
    }
    size_t GetFrontSize() const
    {
        if(buffers.empty()) return 0;
        return buffers.front().size() - frontptr;
    }
    const char *GetFrontPtr() const
    {
        return &buffers.front()[0] + frontptr;
    }
    
    bool Empty() const
    {
        return GetFrontSize() == 0;
    }
    
    size_t GetBufSize() const
    {
        return Empty()
             ? 0
             : buffers.front().size() - frontptr
             + buffers.back().size() * (buffers.size() - 1);
    }
};

static void BlockSignals()
{
    sigset_t set;
    sigemptyset(&set);
    sigaddset(&set, SIGALRM);
    sigprocmask(SIG_BLOCK, &set, NULL);
}

static void UnblockSignals()
{
    sigset_t set;
    sigemptyset(&set);
    sigaddset(&set, SIGALRM);
    sigprocmask(SIG_UNBLOCK, &set, NULL);
}

static std::map<int, buffer> buffers;
static std::map<int, FILE*> fileps;

static bool UpdateBuffer(int fd, buffer& buf)
{
TryAgain:;
    size_t count = buf.GetFrontSize();
    if(!count) return true;

    const char *p = buf.GetFrontPtr();

    ssize_t n = write(fd, p, count);
    if(n < 0 && errno == EINTR) goto TryAgain;
    if(n < 0 && errno == EAGAIN) return true;
    if(n < 0)
    {
        perror("write");
        return false;
    }
    if(n <= 0) return false;

    written_bytes += n;

    buf.PopFront(n);
    goto TryAgain;
}

void AddWriteTask(FILE *fp, const BufferType& data)
{
    BlockSignals();
    
    int fd = fileno(fp);
    
    total_bytes += data.size();
    
    std::map<int, buffer>::iterator i;
    
    i = buffers.lower_bound(fd);
    if(i == buffers.end() || i->first != fd)
    {
        fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
        
        buffer tmp = data;
        if(!UpdateBuffer(fd, tmp))
        {
            // TODO: handle write error
        }
        
        if(tmp.Empty())
        {
            BufferEmptied(fp);
            
            UnblockSignals();
            return;
        }
        
        i = buffers.insert(i, make_pair(fd, tmp));
        fileps.insert(make_pair(fd, fp));
    }
    else
    {
        i->second.PushBack(data);
        
        if(!UpdateBuffer(fd, i->second))
        {
            // TODO: handle write error
        }
    }
    
    UnblockSignals();
}

static void DeleteBuffer(std::map<int, buffer>::iterator i)
{
    std::map<int, FILE*>::iterator fpi = fileps.find(i->first);

    BufferEmptied(fpi->second);
    
    buffers.erase(i);
    fileps.erase(fpi);
}

void UpdateBuffers()
{
    std::map<int, buffer>::iterator i, j;
    for(i = buffers.begin(); i != buffers.end(); i=j)
    {
        j=i; ++j;
        
        UpdateBuffer(i->first, i->second);
        
        if(i->second.Empty())
        {
            DeleteBuffer(i);
        }
    }
}

void FlushBuffers()
{
    std::map<int, buffer>::iterator i,j;
    for(i = buffers.begin(); i != buffers.end(); i=j)
    {
        j=i; ++j;
        
        int fd = i->first;

        fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) &~ O_NONBLOCK);

        buffer& buf = i->second;
        
        for(;;)
        {
            size_t count = buf.GetFrontSize();
            if(!count) break;

            const char *p = buf.GetFrontPtr();

            ssize_t n = write(fd, p, count);
            if(n > 0)
            {
                buf.PopFront(n);
            }
            if(n < 0 && (errno == EAGAIN || errno == EINTR))
                continue;
            if(n < 0) perror("write");
            if(n <= 0) break;
        }

        DeleteBuffer(i);
    }
}

const string GetBufferStatus()
{
    size_t num_bytes = 0;
    
    std::map<int, buffer>::const_iterator i;
    for(i = buffers.begin(); i != buffers.end(); ++i)
    {
        num_bytes += i->second.GetBufSize();
    }
    
    char Buf[64];
    sprintf(Buf, "%zuk/%lluM", num_bytes / 1024, total_bytes/1048576llu);
    return Buf;
}

bool BufferElapse(double nseconds)
{
    std::vector<pollfd> polls;

    std::map<int, buffer>::iterator i;
    for(i = buffers.begin(); i != buffers.end(); ++i)
    {
        if(i->second.Empty()) continue;
        
        pollfd tmp = { i->first, POLLOUT, 0 };
        polls.push_back(tmp);
    }
    
    if(polls.empty())
    {
        usleep((unsigned)(nseconds * 1e6));
        return false;
    }
    
    int timeout = (int)(nseconds * 1000);
    if(!timeout) timeout=1;
    
    int res = poll(&polls[0], polls.size(), timeout);
    if(!res) return true;
    
    /* Block the timer signal while writing so that integrity won't fail */
    BlockSignals();
    
    for(size_t a=0; a<polls.size(); ++a)
    {
        if(polls[a].revents & POLLOUT)
        {
            int fd = polls[a].fd;

            i = buffers.find(fd);
            UpdateBuffer(fd, i->second);
            
            if(i->second.Empty())
            {
                DeleteBuffer(i);
            }
        }
        if(polls[a].revents & POLLERR)
        {
            int fd = polls[a].fd;
            // TODO
            buffers.erase(fd);
        }
    }
    
    UnblockSignals();

    return false;
}

WaitInputRes BufferWaitInput(bool input_eof)
{
    for(;;)
    {
        std::vector<pollfd> polls;

        std::map<int, buffer>::iterator i;
        for(i = buffers.begin(); i != buffers.end(); ++i)
        {
            if(i->second.Empty()) continue;
            
            pollfd tmp = { i->first, POLLOUT, 0 };
            polls.push_back(tmp);
        }
        
        unsigned long long stored_bytes = total_bytes - written_bytes;
        
        if(input_eof && stored_bytes == 0)
        {
            break;
        }
        
        if(!input_eof && stored_bytes < MaxBufferSize) // 100 MB
        {
            pollfd tmp = { STDIN_FILENO, POLLIN, 0 };
            polls.push_back(tmp);
        }
        
        int status = poll(&polls[0], polls.size(), -1);
        if(status < 0) break; // error or interruption

        /* Block the timer signal while writing so that integrity won't fail */
        BlockSignals();
        
        for(size_t a=0; a<polls.size(); ++a)
        {
            if(polls[a].revents & POLLOUT)
            {
                int fd = polls[a].fd;
                
                i = buffers.find(fd);
                UpdateBuffer(fd, i->second);
                
                if(i->second.Empty())
                {
                    DeleteBuffer(i);
                }
            }
            if(polls[a].revents & (POLLIN | POLLHUP))
            {
                return {true,false};
            }
            if(polls[a].revents & POLLERR)
            {
                // An output stream was blocked
                int fd = polls[a].fd;
                buffers.erase(fd);
                return {false,true};
            }
        }
        
        UnblockSignals();
    }
    return {false,false};
}

BufferType::BufferType(size_t size, char c): data()
{
    data = new BufferData;
    data->data = Allocate(size);
    data->size = size;
    std::fill_n(data->data, size, c);
}
BufferType::BufferType(size_t size): data()
{
    data = new BufferData;
    data->data = Allocate(size);
    data->size = size;
}
BufferType::BufferType(const char*b, const char*e): data()
{
    size_t size = e-b;
    data = new BufferData;
    data->data = Allocate(size);
    data->size = size;
    std::copy(b,e, data->data);
}
BufferType::BufferType(const unsigned char*b, const unsigned char*e): data()
{
    size_t size = e-b;
    data = new BufferData;
    data->data = Allocate(size);
    data->size = size;
    std::copy(b,e, data->data);
}

BufferData::~BufferData()
{
    Unallocate(data, size);
}
/*
void* BufferData::operator new(size_t size)
{
    return Allocate(size);
}
void BufferData::operator delete(void *obj, size_t size)
{
    Unallocate(static_cast<char*> (obj), size);
}
*/
